/*----------------------------------------------------------------------------
 * U S B  -  K e r n e l
 *----------------------------------------------------------------------------
 * Name:    msc.h
 * Purpose: USB Mass Storage Class Definitions
 * Version: V1.20
 *----------------------------------------------------------------------------
 *      This software is supplied "AS IS" without any warranties, express,
 *      implied or statutory, including but not limited to the implied
 *      warranties of fitness for purpose, satisfactory quality and
 *      noninfringement. Keil extends you a royalty-free right to reproduce
 *      and distribute executable files created using this software for use
 *      on NXP Semiconductors LPC microcontroller devices only. Nothing else 
 *      gives you the right to use this software.
 *
 * Copyright (c) 2009 Keil - An ARM Company. All rights reserved.
 *---------------------------------------------------------------------------*/
/**
@file msc.h
@brief Definice pro mass-storage class.
*/
#ifndef __MSC_H__
#define __MSC_H__


/* MSC Subclass Codes */
#define MSC_SUBCLASS_RBC                0x01
#define MSC_SUBCLASS_SFF8020I_MMC2      0x02
#define MSC_SUBCLASS_QIC157             0x03
#define MSC_SUBCLASS_UFI                0x04
#define MSC_SUBCLASS_SFF8070I           0x05
#define MSC_SUBCLASS_SCSI               0x06

/* MSC Protocol Codes */
#define MSC_PROTOCOL_CBI_INT            0x00
#define MSC_PROTOCOL_CBI_NOINT          0x01
#define MSC_PROTOCOL_BULK_ONLY          0x50


/* MSC Request Codes */
#define MSC_REQUEST_RESET               0xFF
#define MSC_REQUEST_GET_MAX_LUN         0xFE


/* MSC Bulk-only Stage */
typedef enum _MSC_BulkStages {
  MSC_BS_CBW                    =  0,         // Command Block Wrapper
  MSC_BS_DATA_OUT,                            // Data Out Phase
  MSC_BS_DATA_IN,                             // Data In Phase
  MSC_BS_DATA_IN_LAST,                        // Data In Last Phase
  MSC_BS_DATA_IN_LAST_STALL,                  // Data In Last Phase with Stall
  MSC_BS_CSW,                                 // Command Status Wrapper
  MSC_BS_ERROR                                // Error
} MSC_BulkStages;

/**
@dot
digraph finite_state_machine {
  rankdir=LR;
  size="8,5"
  node [shape = doublecircle]; MSC_BS_CBW;
  node [shape = circle];
  MSC_BS_CBW -> MSC_BS_ERROR [ label = "error" ];
  MSC_BS_CBW -> MSC_BS_CSW   [ label = "fail" ];
  MSC_BS_CSW -> MSC_BS_CBW   [ label = "" ];
  MSC_BS_CBW -> MSC_BS_DATA_IN  [ label = "request read" ];
  MSC_BS_CBW -> MSC_BS_DATA_OUT [ label = "request write, verify" ];
  MSC_BS_DATA_IN  -> MSC_BS_DATA_IN_LAST_STALL [ label = "out of disk" ];
  MSC_BS_DATA_IN  -> MSC_BS_DATA_IN_LAST       [ label = "read end" ];
  MSC_BS_DATA_OUT -> MSC_BS_CSW [ label = "write, verify end" ];
  MSC_BS_DATA_IN_LAST ->       MSC_BS_CSW [ label = "read end" ];
  MSC_BS_DATA_IN_LAST_STALL -> MSC_BS_CSW [ label = "read error" ];
}
@enddot
*/
/* Bulk-only Command Block Wrapper */
typedef struct _MSC_CBW {
  uint32_t dSignature;
  uint32_t dTag;
  uint32_t dDataLength;
  uint8_t  bmFlags;
  uint8_t  bLUN;
  uint8_t  bCBLength;
  uint8_t  CB[16];
} __attribute__((packed)) MSC_CBW;

/* Bulk-only Command Status Wrapper */
typedef  struct _MSC_CSW {
  uint32_t dSignature;
  uint32_t dTag;
  uint32_t dDataResidue;
  uint8_t  bStatus;
} __attribute__((packed)) MSC_CSW;

#define MSC_CBW_Signature               0x43425355
#define MSC_CSW_Signature               0x53425355


/* CSW Status Definitions */
#define CSW_CMD_PASSED                  0x00
#define CSW_CMD_FAILED                  0x01
#define CSW_PHASE_ERROR                 0x02


/* SCSI Commands */
#define SCSI_TEST_UNIT_READY            0x00
#define SCSI_REQUEST_SENSE              0x03
#define SCSI_FORMAT_UNIT                0x04
#define SCSI_INQUIRY                    0x12
#define SCSI_MODE_SELECT6               0x15
#define SCSI_MODE_SENSE6                0x1A
#define SCSI_START_STOP_UNIT            0x1B
#define SCSI_MEDIA_REMOVAL              0x1E
#define SCSI_READ_FORMAT_CAPACITIES     0x23
#define SCSI_READ_CAPACITY              0x25
#define SCSI_READ10                     0x28
#define SCSI_WRITE10                    0x2A
#define SCSI_VERIFY10                   0x2F
#define SCSI_READ12                     0xA8
#define SCSI_WRITE12                    0xAA
#define SCSI_MODE_SELECT10              0x55
#define SCSI_MODE_SENSE10               0x5A


#endif  /* __MSC_H__ */
